/*
 [y,previewEstimate,previewPartition] = CTFPlateletApprox(x[,pen,pen_w,previewBoxSidelength,noiseType]);
 */
#include <math.h>
#include <stdio.h>
#include "mex.h"
#include "matrix.h"
#include <stdlib.h>
#define max(A,B) (A > B ? A : B)
#define min(A,B) (A < B ? A : B)
#define even(x)  ((x & 1) ? 0 : 1)
#define isint(x) ((x - floor(x)) > 0.0 ? 0 : 1)
#define mat(a, i, j) (*(a + (m*(j)+i)))

int mod(int A, int B){
  int y;
  double a = (double)A;
  double b = (double)B;
  y = (int)((a/b - floor(a/b))*b);
  return y;
}

void fitPlatelet(double **obs, double **plateFit, int oi, int oj, int m, int n) {
  char *trans = "N";
  double *T, *details, *obs_1D, *plateFit_1D, *work, Tm, Tn;
  int i, j, lwork, k = 0;
  int M, N, NRHS, LDA, LDB, INFO, INCX, INCY;
  double ALPHA, BETA;
  T = mxCalloc(m*n*3,sizeof(double));
  details = mxCalloc(m*n*3,sizeof(double));
  obs_1D = mxCalloc(m*n,sizeof(double));
  plateFit_1D = mxCalloc(m*n,sizeof(double));
  lwork = m*n*2;
  work = mxCalloc(lwork,sizeof(double));
  k =0;
  for (i = 0; i < m; i++)  {
    for (j = 0; j < n; j++)    {
      T[k] = 1.0;
      T[k+m*n] = (double)i;
      T[k+m*n*2] = (double)j;
      details[k] = 1.0;
      details[k+m*n] = (double)i;
      details[k+m*n*2] = (double)j;
      obs_1D[k] = obs[i+oi][j+oj];
      k++;
    }
  }
  M = k;
  N = 3;
  NRHS = 1;
  LDA = M;
  LDB = M;
  ALPHA = 1.0;
  INCX = 1;
  BETA = 0.0;
  INCY = 1;
  dgels_(trans,&M,&N,&NRHS,details,&LDA,obs_1D,&LDB,work,&lwork,&INFO);
  dgemv_(trans,&M,&N,&ALPHA,T,&M,obs_1D,&INCX,&BETA,plateFit_1D,&INCY);
  k = 0;
  for (i = 0; i < m; i++)  {
    for (j = 0; j < n; j++)    {
      plateFit[i+oi][j+oj] = plateFit_1D[k];
      k++;
    }
  }
  mxFree(T);
  mxFree(details);
  mxFree(obs_1D);
  mxFree(plateFit_1D);
  mxFree(work);
}


void initPlatelets(int sh, double **y, double **z, int m, int n)
{
  int i, j, k, kx, ky = 0;
  for (i = 0; i < m; i+=sh) {
    for (j = 0; j < n; j+=sh) {
      fitPlatelet(y,z,i,j,sh,sh);
    }
  }
}

void plateletApprox(double **x, int m, int n, double pen, double **y,
                    double **dispPartition, double **c, char *noiseType)
{
  int L, im, in, iL, j, k, dyadLen;
  double pMerge,ptmp,pSplit;
  double **merge, **partition, **corners, **cx, **cy;
	if (n==1){
		n = m;
		m = 1;
	}
	L = (int) floor(log(n)/log((double)2));
	merge = (double **)mxCalloc(m,sizeof(double *));
	partition = (double **)mxCalloc(m,sizeof(double *));
	corners = (double **)mxCalloc(m,sizeof(double *));
	cx = (double **)mxCalloc(m,sizeof(double *));
	cy = (double **)mxCalloc(m,sizeof(double *));
	for (im = 0; im < m; im++) {
		merge[im] = (double *)mxCalloc(n,sizeof(double));
		partition[im] = (double *)mxCalloc(n,sizeof(double));
		corners[im] = (double *)mxCalloc(n,sizeof(double));
		cx[im] = (double *)mxCalloc(n,sizeof(double));
		cy[im] = (double *)mxCalloc(n,sizeof(double));
		for (in=0;in<n;in++) {
			y[im][in] = x[im][in];
			partition[im][in] = 1.0;
			dispPartition[im][in] = rand();
			corners[im][in] = 1.0;
			cx[im][in] = im;
			cy[im][in] = in;
		}
	}
	for (iL=0;iL<L;iL++) {
		dyadLen = (int)pow(2.0,iL+1);
		for (im=0;im<m;im+=dyadLen){
			for (in=0;in<n;in+=dyadLen) {
				ptmp = 0;
				for (j=0;j<dyadLen;j++) {
					for (k=0;k<dyadLen;k++){
						/*merge[im][in] += x[im+j][in+k]/dyadLen/dyadLen;*/
						ptmp += dispPartition[im+j][in+k]/dyadLen/dyadLen;
					}
				}
				fitPlatelet(x,merge,im,in,dyadLen,dyadLen);
				pMerge = 0;
				pSplit = 0;
				for (j=0;j<dyadLen;j++) {
					for (k=0;k<dyadLen;k++) {
						if (strcmp(noiseType,"Poisson")==0) {
							if ((merge[im+j][in+k]==0.0)&&(x[im+j][in+k]==0.0)) {
								pMerge -= pen/dyadLen/dyadLen;
							}
							else {
								pMerge += (-merge[im+j][in+k] + x[im+j][in+k]*log(merge[im+j][in+k])) - pen/dyadLen/dyadLen;
							}
							if ((y[im+j][in+k]==0.0)&&(x[im+j][in+k]==0.0)) {
								pMerge -= partition[im+j][in+k];
							}
							else {
								pSplit += (-y[im+j][in+k] + x[im+j][in+k]*log(y[im+j][in+k])) - partition[im+j][in+k]*pen;
							}
						}
						else {
							pMerge += -pow(x[im+j][in+k]-merge[im+j][in+k],2.0) - pen/dyadLen/dyadLen;
							pSplit += -pow(x[im+j][in+k]-y[im+j][in+k],2.0) - partition[im+j][in+k]*pen;
						}
					}
				}
				if (pMerge >= pSplit) {
					for (j=0;j<dyadLen;j++){
						for (k=0;k<dyadLen;k++) {
							y[im+j][in+k] = merge[im+j][in+k];
							partition[im+j][in+k] = 1.0/dyadLen/dyadLen;
							dispPartition[im+j][in+k] = ptmp;
							corners[im+j][in+k] = dyadLen;
							cx[im+j][in+k] = im;
							cy[im+j][in+k] = in;
						}
					}
				}
			}
		}
	}
	for (im=0;im<m;im++) {
		for (in=0;in<n;in++){
			c[im*n+in][0] = cx[im][in];
			c[im*n+in][1] = cy[im][in];
			c[im*n+in][2] = corners[im][in];
		}
	}
	for (im = 0; im < m; im++) {
		mxFree(merge[im]);
		mxFree(partition[im]);
		mxFree(corners[im]);
		mxFree(cx[im]);
		mxFree(cy[im]);
	}
	mxFree(merge);
	mxFree(partition);
	mxFree(corners);
	mxFree(cx);
	mxFree(cy);
}


void shimmey(double **x, double **y, int s, int m, int n)
{
	int i, j;
  /* y[i,j] = x[i+1,j+1] */
  for (i = 0; i<m; i++) {
    for (j=0; j<n; j++) {
      y[i][j] = x[mod(i-s,m)][mod(j-s,n)];
    }
  }
}
/* subroutine for filling up data */
void assign( double *pl, double *pr, int n )
{
  int i;
  for (i=0; i < n; i++)
    pl[i]=pr[i];
}

void getMask(double** mask, double r, double th, int N){
  double thr; 
  double th2; 
  double th2r;
  double x,y,R;
  int i,j;
  double b1, b2,b3,b4;
  thr = th*M_PI/180;
  th2 = fmod(th, 90.0);
  if (th2 >= 45){
    th2 = 90 - th2;
  }
  th2r = th2*M_PI/180;
  b1 = -.5*(cos(th2r)+sin(th2r));
  b2 = -.5*(cos(th2r)-sin(th2r));
  b3 = .5*(cos(th2r)-sin(th2r));
  b4 = .5*(cos(th2r)+sin(th2r));
  x = -N/2 - .5;
  for (i=0;i<N;i++){
    x = x + 1;
    y = N/2 + .5;
    for (j=0;j<N;j++){
      y = y - 1;
      R = r - x*cos(thr)-y*sin(thr);
      if (R < b1){
        mask[i][j] = 0;
      }
      else if ((R >= b1) && (R < b2)){
        mask[i][j] = .25 + ((R*R + R*(cos(th2r) + sin(th2r)) + .25)/(2*cos(th2r)*sin(th2r)));
      }
      else if ((R >= b2) && (R < b3)){
        mask[i][j] = .5 + R/cos(th2r);
      }
      else if ((R >= b3) && (R < b4)){
        mask[i][j] = .75 - tan(th2r)/4 - 1/(4*tan(th2r)) + ((-1*R*R + R*(cos(th2r) + sin(th2r)) + .25)/(2*cos(th2r)*sin(th2r)));
      }
      else{
				mask[i][j] = 1;
			}
			mask[i][j] = (mask[i][j]>0.5);
		}
	}
}
double err(double **data, double **est,int m, int n,char *noiseType)
{
  double error = 0;
  int i,j;
  if (strcmp(noiseType,"Poisson")==0) {
    for (i = 0; i < m; i++) {
      for (j = 0; j < n; j++) {
        if ((est[i][j]==0.0) && (data[i][j]==0.0)) {
          error += 0.0;
        }
        else {
          error =  error+est[i][j] - data[i][j]*log(est[i][j]);
        }
      }
    }
  }
  else if (strcmp(noiseType,"Gaussian")==0)
  {
    for (i = 0; i < m; i++) {
      for (j = 0; j < n; j++) {
        error += pow(est[i][j] - data[i][j], 2.0);
      }
    }
  }
  else
    mexErrMsgTxt("noiseType must be either 'Poisson' or 'Gaussian'.");
  return error;
}

void fitMaskedPlatelet(double **obs, double **plateFit, double **mask, int oi, int oj, int m, int n) {
  char *trans = "N";
  double *T, *details, *obs_1D, *plateFit_1D, *work, Tm, Tn;
  int i, j, lwork, k = 0;
  int M, N, NRHS, LDA, LDB, INFO, INCX, INCY;
  double ALPHA, BETA;
  T = mxCalloc(m*n*6,sizeof(double));
  details = mxCalloc(m*n*6,sizeof(double));
  obs_1D = mxCalloc(m*n,sizeof(double));
  plateFit_1D = mxCalloc(m*n,sizeof(double));
  lwork = m*n*2;
  work = mxCalloc(lwork,sizeof(double));
  k =0;
  for (i = 0; i < m; i++)  {
    for (j = 0; j < n; j++)    {
      T[k] = 1.0*mask[i][j];
      T[k+m*n] = (double)i*mask[i][j];
      T[k+m*n*2] = (double)j*mask[i][j];
      T[k+m*n*3] = 1.0*(1-mask[i][j]);
      T[k+m*n*4] = (double)i*(1-mask[i][j]);
      T[k+m*n*5] = (double)j*(1-mask[i][j]);
      details[k] = 1.0*mask[i][j];
      details[k+m*n] = (double)i*mask[i][j];
      details[k+m*n*2] = (double)j*mask[i][j];
      details[k+m*n*3] = 1.0*(1-mask[i][j]);
      details[k+m*n*4] = (double)i*(1-mask[i][j]);
      details[k+m*n*5] = (double)j*(1-mask[i][j]);
      obs_1D[k] = obs[i+oi][j+oj];
      k++;
    }
  }
  M = k;
  N = 6;
  NRHS = 1;
  LDA = M;
  LDB = M;
  ALPHA = 1.0;
  INCX = 1;
  BETA = 0.0;
  INCY = 1;
  dgels_(trans,&M,&N,&NRHS,details,&LDA,obs_1D,&LDB,work,&lwork,&INFO);
  dgemv_(trans,&M,&N,&ALPHA,T,&M,obs_1D,&INCX,&BETA,plateFit_1D,&INCY);
  k = 0;
  for (i = 0; i < m; i++)  {
    for (j = 0; j < n; j++)    {
      plateFit[i+oi][j+oj] = plateFit_1D[k];
      k++;
    }
  }
  mxFree(T);
  mxFree(details);
  mxFree(obs_1D);
  mxFree(plateFit_1D);
  mxFree(work);
}

void fitWedgePlate(double** I, int DIM, double** OUTIMAGE, double** OUTMASK, int num_r, char *noiseType){
  int i,j,s,s2 = 0;
  double **mask, **wedgePlateEst;
  double r, th, WC1, WC2, error;
  double bestError, bestR, bestTH = 0;
  mask = (double**)mxCalloc(DIM,sizeof(double*));
  wedgePlateEst = (double**)mxCalloc(DIM,sizeof(double*));
  for (i = 0; i < DIM; i++){
    mask[i] = (double*)mxCalloc(DIM,sizeof(double));
    wedgePlateEst[i] = (double*)mxCalloc(DIM,sizeof(double));
  }
  bestError = 100000000;
  for (s=0;s<num_r;s++){
    for (s2=0;s2<num_r;s2++){
      r = (double)DIM/sqrt(2.0)*(double)s/(double)(num_r+1.0);
      th = 360.0*(double)s2/(double)(num_r+1.0);
      getMask(mask,r,th,DIM);
			
      fitMaskedPlatelet(I,wedgePlateEst,mask,0,0,DIM,DIM);
			
      error = err(I,wedgePlateEst,DIM,DIM,noiseType);
      if (error < bestError) {
        bestError = error;
        bestR = r;
        bestTH = th;
        for (i = 0; i < DIM; i++){
          for (j = 0; j < DIM; j++){
            OUTIMAGE[i][j] = wedgePlateEst[i][j];
            OUTMASK[i][j] = mask[i][j];
          }
        }
      }
    }
  }
  for (i = 0; i < DIM; i++){
    mxFree(mask[i]);
    mxFree(wedgePlateEst[i]);
  }
  mxFree(mask);
  mxFree(wedgePlateEst);
}	

void refineEst(double **est, double **data, double **preview,
               double pen, double pen_w, int m,
               int n, int smallestBoxSidelength, double **partitionCorners,
               int numSmallBoxes, char *noiseType, int numWedges)
{
  int i,j,k,cj,ck,cl;
  double **dataBlock, **plateEst, **wedgePlateEst, **wedgePlateMask, **data2, **est2;
  double meanVal;
  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++) {
      est[i][j] = preview[i][j];
    }
  }
  dataBlock = (double**)mxCalloc(smallestBoxSidelength,sizeof(double*));
  wedgePlateEst = (double**)mxCalloc(smallestBoxSidelength,sizeof(double*));
  wedgePlateMask = (double**)mxCalloc(smallestBoxSidelength,sizeof(double*));
  plateEst = (double**)mxCalloc(smallestBoxSidelength,sizeof(double*));
  for (i = 0; i < smallestBoxSidelength; i++) {
    dataBlock[i]= (double *)mxCalloc(smallestBoxSidelength,sizeof(double));
    wedgePlateEst[i] = (double *)mxCalloc(smallestBoxSidelength,sizeof(double));
    wedgePlateMask[i] = (double *)mxCalloc(smallestBoxSidelength,sizeof(double));
    plateEst[i] = (double *)mxCalloc(smallestBoxSidelength,sizeof(double));
  }
  for (i = 0; i < numSmallBoxes; i++) {
    if ( partitionCorners[i][2] == smallestBoxSidelength) {
      cj = partitionCorners[i][0];
      ck = partitionCorners[i][1];
      cl = partitionCorners[i][2];
      meanVal = 0;
      for (j = 0; j < cl; j++) {
        for (k = 0; k < cl; k++) {
          dataBlock[j][k] = data[j+cj][k+ck];
          meanVal += dataBlock[j][k]/smallestBoxSidelength/smallestBoxSidelength;
        }
      }
			
      fitPlatelet(dataBlock,plateEst,0,0,cl,cl);
      
      fitWedgePlate(dataBlock, smallestBoxSidelength, wedgePlateEst, wedgePlateMask, numWedges ,noiseType);
			
      if ((err(dataBlock,plateEst,smallestBoxSidelength,smallestBoxSidelength,noiseType)+pen) <
          (err(dataBlock,wedgePlateEst,smallestBoxSidelength,smallestBoxSidelength,noiseType)
           + 2.0*pen_w)) {
        for (j = 0; j < cl; j++) {
          for (k = 0; k < cl; k++) {
            est[j+cj][k+ck] = plateEst[j][k];
          }
        }
      }
      else {
        for (j = 0; j < cl; j++) {
          for (k = 0; k < cl; k++) {
            est[j+cj][k+ck] = wedgePlateEst[j][k];
          }
        }
      }      
    }
  }
  for (i = 0; i < smallestBoxSidelength; i++) {
    mxFree(dataBlock[i]);
    mxFree(wedgePlateEst[i]);
    mxFree(wedgePlateMask[i]);
    mxFree(plateEst[i]);
  }
  mxFree(dataBlock);
  mxFree(wedgePlateEst);
  mxFree(wedgePlateMask);
  mxFree(plateEst);
}

/******************************************/
/* main																		*/
/******************************************/
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
  /******************************************/
  /* call in MATLAB:
  [refinedEst, previewEst, previewPartition, corners, shimmeyPreviewEst, shimmeyPreviewPartition] =
	CTFWedgeletApprox(x,previewBoxSidelength,noiseType,pen,pen_w)
	/******************************************/
	
	double *xCol, *yCol, *y2Col, *yPartitionCol, *y2PartitionCol,*zCol;
	double **x, **xshimmey, **y, **y2, **z;
	double **yPartition, **y2Partition, *y12Corners;
	double **yshimmey, **yshimmeyPartition, **y1Corners, **y2Corners;
	double **corners, *cornersCol, **coarsePlatelets;
	int m, n, i, j, L, numWedges;
	double mtest, ntest, pen, pen_w, pen_preview, sum = 0;
	int previewBoxSidelength, numSmallBoxes;
	char *noiseType;
	mxArray *uniqueArgs[2], *uniqueCorners[1];
	/******************************************/
	/* check for correct # of input variables */
	/******************************************/
	if (nrhs>6){
		mexErrMsgTxt("There are at most 6 input parameters allowed!");
		return;
	}
	if (nrhs<1){
		mexErrMsgTxt("There is at least 1 input parameter required!");
		return;
	}
	/**************************************/
	/* get data */
	/**************************************/
	xCol = mxGetPr(prhs[0]);
	n = mxGetN(prhs[0]);
	m = mxGetM(prhs[0]);
	/**************************************/
	/* Check the ROW dimension of input */
	/**************************************/
	if ((m < 2)||(n<2))
		mexErrMsgTxt("This is an image processing method; it cannot process a vector.");
	if(m > 1){
		mtest = (double) log(m)/log((double)2);
		if (!isint(mtest))
			mexErrMsgTxt("The matrix row dimension must be of size m*2^(L)");
	}
	/**************************************/
	/* Check the COLUMN dimension of input  */
	/**************************************/
	if(n > 1){
		ntest = (double) log(n)/log(2.0);
		if (!isint(ntest))
			mexErrMsgTxt("The matrix column dimension must be of size n*2^(L)");
	}	
	/**************************************/
	/* allocate memory */
	/**************************************/
	x = (double **)mxCalloc(m,sizeof(double *));
	xshimmey = (double **)mxCalloc(m,sizeof(double *));
	y = (double **)mxCalloc(m,sizeof(double *));
	yPartition = (double **)mxCalloc(m,sizeof(double *));
	y2 = (double **)mxCalloc(m,sizeof(double *));
	y2Partition = (double **)mxCalloc(m,sizeof(double *));
	yshimmey = (double **)mxCalloc(m,sizeof(double *));
	yshimmeyPartition = (double **)mxCalloc(m,sizeof(double *));
	z = (double **)mxCalloc(m,sizeof(double *));
	coarsePlatelets = (double **)mxCalloc(m,sizeof(double*));
	for (i = 0; i < m; i++) {
		x[i] = (double *)mxCalloc(n,sizeof(double));
		xshimmey[i] = (double *)mxCalloc(n,sizeof(double));
		y[i] = (double *)mxCalloc(n,sizeof(double));
		yPartition[i] = (double *)mxCalloc(n,sizeof(double));
		yshimmey[i] = (double *)mxCalloc(n,sizeof(double));
		yshimmeyPartition[i] = (double *)mxCalloc(n,sizeof(double));
		y2[i] = (double *)mxCalloc(n,sizeof(double));
		y2Partition[i] = (double *)mxCalloc(n,sizeof(double));
		z[i] = (double *)mxCalloc(n,sizeof(double));
		coarsePlatelets[i] = (double *)mxCalloc(n,sizeof(double));
	}
	y1Corners = (double **)mxCalloc(m*n,sizeof(double *));
	y2Corners = (double **)mxCalloc(m*n,sizeof(double *));
	for (i = 0; i < m*n; i++) {
		y1Corners[i] = (double *)mxCalloc(3,sizeof(double));
		y2Corners[i] = (double *)mxCalloc(3,sizeof(double));
	}
	y12Corners = (double *)mxCalloc(m*n*6,sizeof(double));
	
	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			x[i][j] = xCol[i*n+j];
		}
	}
	
	/**************************************/
	/* set previewBoxSidelength */
	/**************************************/
	if ((nrhs < 2) || mxGetN(prhs[1])==0) {
		previewBoxSidelength = (int) pow(2.0,floor(log2(sqrt(pow((double)n*m,1.0/3.0)))));
	}
	else {
		previewBoxSidelength = (int) mxGetScalar(prhs[1]);
	}
	
	/**************************************/
	/* set noiseType */
	/**************************************/
	if (nrhs < 3) {
		noiseType = "Gaussian";
	}
	else {
		if (mxIsChar(prhs[2]) != 1) {
			mexErrMsgTxt("noiseType must be a string");
		}
		i = (mxGetM(prhs[2])*mxGetN(prhs[2])) + 1;
		noiseType = mxCalloc(i,sizeof(char));
		mxGetString(prhs[2],noiseType,i);
	}
	if ((strcmp(noiseType,"Poisson") != 0) && (strcmp(noiseType,"Gaussian")!=0)) {
		mexErrMsgTxt("noiseType must be either 'Poisson' or 'Gaussian'.");
	}
	/**************************************/
	/*
	 set penalty
	 Gaussian
	 pen = 2*noise_var*((1/3)*log(n));
	 Poisson
	 pen = (1/3)log(n)
	 */
	/**************************************/
	if ((nrhs < 4)||(mxGetN(prhs[3])==0)) {
		if (strcmp(noiseType,"Poisson")==0) {
			for (i = 0; i < n*m; i++) {
				sum += xCol[i];
			}
			if (sum < 0) {
				mexErrMsgTxt("Poisson data cannot be negative.");
			}
			pen = (double) log(sum)*3.0/3.0;
			pen_w = (double) log(sum)*4.0/3.0;
		}
		else if (strcmp(noiseType,"Gaussian")==0) {
			pen = (double) log(m*n)*3.0/3.0;
			pen_w = (double) log(m*n)*4.0/3.0;
		}
		else {
			printf("noiseType = %s\n",noiseType);
			mexErrMsgTxt("noiseType must be either 'Poisson' or 'Gaussian'.");
		}
	}
	else {
		pen = (double) mxGetScalar(prhs[3]);
		pen_w = (double) mxGetScalar(prhs[4]);
	}
	if ((pen < 0)||(pen_w < 0)) {
		mexErrMsgTxt("The penalty must be positive.");
	}
	pen_preview = pen*2.0/3.0;
	
	/**************************************/
	/* set numWedges */
	/**************************************/
	if ((nrhs < 6) || mxGetN(prhs[5])==0) {
		numWedges = (int) floor(pow(n*m,1.0/3.0));
	}
	else {
		numWedges = (int) floor(sqrt(mxGetScalar(prhs[5])));
	}
	/**************************************/
	/* set up outputs */
	/**************************************/
	
	/* estimate */
	plhs[0] = mxCreateDoubleMatrix(m,n,mxREAL);
	zCol = mxGetPr(plhs[0]);
	
	/* preview est */
	plhs[1] = mxCreateDoubleMatrix(m,n,mxREAL);
	yCol = mxGetPr(plhs[1]);
	
	/* preview partition */
	plhs[2] = mxCreateDoubleMatrix(m,n,mxREAL);
	yPartitionCol = mxGetPr(plhs[2]);
	
	/* corners */
	plhs[3] = mxCreateDoubleMatrix(1,1,mxREAL);
	
	/* shimmeyPreview Estimate */
	plhs[4] = mxCreateDoubleMatrix(m,n,mxREAL);
	y2Col = mxGetPr(plhs[4]);
	
	/* shimmeyPreview Partition */
	plhs[5] = mxCreateDoubleMatrix(m,n,mxREAL);
	y2PartitionCol = mxGetPr(plhs[5]);
	
	/**************************************/
	/* get the preview estimate */
	/**************************************/
	
	initPlatelets(previewBoxSidelength,x,coarsePlatelets,m,n);
	plateletApprox(coarsePlatelets, m, n, pen_preview, y, yPartition,
								 y1Corners, noiseType);
	/**************************************/
	/* get the shimmey preview estimate */
	/**************************************/
	
	shimmey(coarsePlatelets,xshimmey,previewBoxSidelength,m,n);
	plateletApprox(xshimmey, m, n, pen_preview, yshimmey, yshimmeyPartition,
								 y2Corners, noiseType);
	/**************************************/
	/* undo the shimmey on the shimmey preview estimate */
	/**************************************/
	shimmey(yshimmey,y2,-previewBoxSidelength,m,n);
	shimmey(yshimmeyPartition,y2Partition,-previewBoxSidelength,m,n);
	
	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			y[i][j] = (y[i][j]+y2[i][j])/2.0;
		}
	}
	
	/**************************************/
	/* get the partition corners */
	/**************************************/
	for (i = 0; i < m*n; i++) {
		y12Corners[i+0*m*n] = y1Corners[i][0];
		y12Corners[i+2*m*n] = y1Corners[i][1];
		y12Corners[i+4*m*n] = y1Corners[i][2];
		y12Corners[i+1*m*n] = mod(y2Corners[i][0]-previewBoxSidelength,m);
		y12Corners[i+3*m*n] = mod(y2Corners[i][1]-previewBoxSidelength,n);
		y12Corners[i+5*m*n] = y2Corners[i][2];
	}
	
	uniqueArgs[0] = mxCreateDoubleMatrix(m*n*2,3,mxREAL);
	assign(mxGetPr(uniqueArgs[0]),y12Corners,m*n*6);
	uniqueArgs[1] = mxCreateString("rows");
	mexCallMATLAB(1,uniqueCorners,2,uniqueArgs,"unique");
	plhs[3] = uniqueCorners[0];
	cornersCol = mxGetPr(uniqueCorners[0]);
	numSmallBoxes = mxGetM(uniqueCorners[0]);
	
	corners = (double **)mxCalloc(numSmallBoxes,sizeof(double *));
	
	for (i = 0; i < numSmallBoxes; i++) {
		corners[i] = (double *)mxCalloc(3,sizeof(double));
		for (j = 0; j < 3; j++) {
			corners[i][j] = cornersCol[i+j*numSmallBoxes];
		}
	}
	
	/**************************************/
	/* refine the estimate */
	/**************************************/
	refineEst(z, x, y, pen, pen_w, m, n, previewBoxSidelength, corners,
						numSmallBoxes, noiseType, numWedges);
	/**************************************/
	/* proper output format */
	/**************************************/
	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			zCol[i*n+j] = z[i][j];
			yCol[i*n+j] = y[i][j];
			y2Col[i*n+j] = y2[i][j];
			yPartitionCol[i*n+j] = yPartition[i][j];
			y2PartitionCol[i*n+j] = y2Partition[i][j];
		}
	}
	
	/**************************************/
	/* return to matlab */
	/**************************************/
}

