/*
 MATLAB description:
 y = haarApprox(x,pen,smallestBoxSidelength);
 */

#include <math.h>
#include <stdio.h>
#include "mex.h"
#include "matrix.h"

#define max(A,B) (A > B ? A : B)
#define min(A,B) (A < B ? A : B)
#define even(x)  ((x & 1) ? 0 : 1)
#define isint(x) ((x - floor(x)) > 0.0 ? 0 : 1)
/*#define mat(a, i, j) (a[m*(j)+i]) */
#define mat(a, i, j) (*(a + (m*(j)+i)))

double mod(double A, double B){
  double y;
  y = ((double)((double)A/(double)B - floor((double)A/(double)B))*(double)B);
  return y;
}

void sum_dyads(int sh, double *y, double *z, int m, int n)
{
  int i,j,kx,ky = 0;
  double z_tmp;
  for (i = 0; i < n; i+=sh) {
    for (j = 0; j < m; j+=sh) {
      z_tmp = 0;
      for (kx = 0; kx < sh; kx++) {
        for (ky = 0; ky < sh; ky++) {
          z_tmp =  z_tmp + *(y+j+i*m+ky+kx*m);
        }
      }
      for (kx = 0; kx < sh; kx++) {
        for (ky = 0; ky < sh; ky++) {
          *(z+j+i*m+ky+kx*m) = z_tmp/sh/sh;
        }
      }
    }
  }
}


haarApprox(double *x, int m, int n, double pen, double *p, double *y,
           double *c)
{
  int L, im, in, iL, j, k, dyadLen;
  double pMerge,ptmp,pSplit;
  double **xShift, **yShift, **merge, **partition, **dispPartition, **corners, **cx, **cy;

  if (n==1){
    n = m;
    m = 1;
  }
  L = (int) floor(log(n)/log((double)2));

  xShift = (double **)mxCalloc(m,sizeof(double *));
  yShift = (double **)mxCalloc(m,sizeof(double *));
  merge = (double **)mxCalloc(m,sizeof(double *));
  partition = (double **)mxCalloc(m,sizeof(double *));
  dispPartition = (double **)mxCalloc(m,sizeof(double *));
  corners = (double **)mxCalloc(m,sizeof(double *));
  cx = (double **)mxCalloc(m,sizeof(double *));
  cy = (double **)mxCalloc(m,sizeof(double *));

  for (im = 0; im < m; im++) {
    *(xShift+im) = (double *)mxCalloc(n,sizeof(double));
    *(yShift+im) = (double *)mxCalloc(n,sizeof(double));
    *(merge+im) = (double *)mxCalloc(n,sizeof(double));
    *(partition+im) = (double *)mxCalloc(n,sizeof(double));
    *(dispPartition+im) = (double *)mxCalloc(n,sizeof(double));
    *(corners+im) = (double *)mxCalloc(n,sizeof(double));
    *(cx+im) = (double *)mxCalloc(n,sizeof(double));
    *(cy+im) = (double *)mxCalloc(n,sizeof(double));
  }

  for (j = 0; j< n*m; j++)
    *(y+j) = 0;

  for (im=0;im<m;im++){
    for (in=0;in<n;in++) {
      *(*(xShift+im)+in) = *(x+im+n*in);
      *(*(yShift+im)+in) = *(*(xShift+im)+in);
      *(*(partition+im)+in) = 1.0;
      *(*(dispPartition+im)+in) = rand();
      *(*(corners+im)+in) = 1.0;
      *(*(cx+im)+in) = im;
      *(*(cy+im)+in) = in;
    }
  }
  
  for (iL=0;iL<L;iL++) {
    dyadLen = (int)pow(2.0,iL+1);
    for (im=0;im<m;im+=dyadLen){
      for (in=0;in<n;in+=dyadLen) {
        *(*(merge+im)+in) = 0;
        ptmp = 0;
        for (j=0;j<dyadLen;j++)
          for (k=0;k<dyadLen;k++){
            *(*(merge+im)+in) += *(*(xShift+im+j)+in+k)/dyadLen/dyadLen;
            ptmp += *(*(dispPartition+im+j)+in+k)/dyadLen/dyadLen;
          }
        pMerge = 0;
        pSplit = 0;
        for (j=0;j<dyadLen;j++) {
          for (k=0;k<dyadLen;k++) {
            *(*(merge+im+j)+in+k) = *(*(merge+im)+in);
            pMerge += -pow(*(*(xShift+im+j)+in+k)-*(*(merge+im+j)+in+k),2.0) - pen/dyadLen/dyadLen;
            pSplit += -pow(*(*(xShift+im+j)+in+k)-*(*(yShift+im+j)+in+k),2.0) - *(*(partition+im+j)+in+k)*pen;
          }
        }
        if (pMerge >= pSplit) {
          for (j=0;j<dyadLen;j++){
            for (k=0;k<dyadLen;k++) {
              *(*(yShift+im+j)+in+k) = *(*(merge+im+j)+in+k);
              *(*(partition+im+j)+in+k) = 1.0/dyadLen/dyadLen;
              *(*(dispPartition+im+j)+in+k) = ptmp;
              *(*(corners+im+j)+in+k) = dyadLen;
              *(*(cx+im+j)+in+k) = im;
              *(*(cy+im+j)+in+k) = in;
            }
          }
        }
      }
    }
  }
  for (im=0;im<m;im++) {
    for (in=0;in<n;in++){
      *(y+im+in*n) += *(*(yShift+im)+in);
      *(p+im+in*n) += *(*(dispPartition+im)+in);
      *(c+2*m*n+im+in*n) = *(*(corners+im)+in);
      *(c+im+in*n) = *(*(cx+im)+in);
      *(c+m*n+im+in*n) = *(*(cy+im)+in);
    }
  }
  for (im=0; im<m; im++) {
    mxFree(xShift[im]);
    mxFree(yShift[im]);
    mxFree(merge[im]);
    mxFree(partition[im]);
    mxFree(dispPartition[im]);
    mxFree(corners[im]);
    mxFree(cx[im]);
    mxFree(cy[im]);
  }
  mxFree(xShift);
  mxFree(yShift);
  mxFree(merge);
  mxFree(partition);
  mxFree(dispPartition);
  mxFree(corners);
  mxFree(cx);
  mxFree(cy);
}

void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
  double *x, *y, *z, *p, *corners;
  int m, n, i, j, L, smallestSidelength;
  double mtest, ntest, pen;
  mxArray *zmat;

  /* check for correct # of input variables */
  if (nrhs>3){
    mexErrMsgTxt("There are at most 3 input parameters allowed!");
    return;
  }
  if (nrhs<1){
    mexErrMsgTxt("There is at least 1 input parameter required!");
    return;
  }

  if (nrhs<2){
    mexErrMsgTxt("At this point the penalty must be input.");
    return;
  }
  else /* Use theoretical penalty */ {
    /* Check to make sure the second input argument is a scalar. */
    if (!mxIsDouble(prhs[1]) || mxIsComplex(prhs[1]) ||
        mxGetN(prhs[1])*mxGetM(prhs[1]) != 1) {
      mexErrMsgTxt("Input pen must be a scalar.");
    }
    pen = (double) mxGetScalar(prhs[1]);
    if (pen < 0)
      mexErrMsgTxt("The penalty must be positive.");
  }

  if (nrhs < 3) {
    smallestSidelength = 1;
  }
  else {
    /* Check to make sure the third input argument is a scalar. */
    if (!mxIsDouble(prhs[2]) || mxIsComplex(prhs[2]) ||
        mxGetN(prhs[2])*mxGetM(prhs[2]) != 1) {
      mexErrMsgTxt("Input sh must be a scalar.");
    }
    smallestSidelength = (int) mxGetScalar(prhs[2]);
  }

  x = mxGetPr(prhs[0]);
  n = mxGetN(prhs[0]);
  m = mxGetM(prhs[0]);

  /* Check image dimensions */
  if ((m>1)&&(n>1)&&(m!=n))
    mexErrMsgTxt("Images must be square.");

  /* Check the ROW dimension of input */
  if(m > 1){
    mtest = (double) log(m)/log((double)2);
    if (!isint(mtest))
      mexErrMsgTxt("The matrix row dimension must be of size m*2^(L)");
  }
  /* Check the COLUMN dimension of input  */
  if(n > 1){
    ntest = (double) log(n)/log(2.0);
    if (!isint(ntest))
      mexErrMsgTxt("The matrix column dimension must be of size n*2^(L)");
  }
  plhs[0] = mxCreateDoubleMatrix(m,n,mxREAL);
  plhs[1] = mxCreateDoubleMatrix(m,n,mxREAL);
  plhs[2] = mxCreateDoubleMatrix(m*n,3,mxREAL);
  y = mxGetPr(plhs[0]);
  p = mxGetPr(plhs[1]);
  corners = mxGetPr(plhs[2]);
  zmat = mxCreateDoubleMatrix(m,n,mxREAL);
  z = mxGetPr(zmat);
  sum_dyads(smallestSidelength,x,z,m,n);
  haarApprox(z, m, n, pen, p, y, corners);
}
